#include <BasicLinearAlgebra.h>

using namespace BLA;


// ---------------------------------------------------------------- //
// Ultrasoninc Sensor
// Re-written by Gentrit Gashi
// Using M5Stack ----
// Adjusted on 5th of November 2022
// ---------------------------------------------------------------- //


// Making this code more efficient ! ///
// DDRB 8 - 13 // DDRD 0 - 7 // Register values
// PORTB 8 - 13 // PORTD 0 - 7 // Sets voltages // 1 for 5V(HIGH) - 0 for 0V(LOW)

/*
#define echo1 B00000100
#define trig1 B00001000 //attach pin D3 Arduino

#define echo2 B00010000 // attach pin D2 Arduino
#define trig2 B00100000 //attach pin D3 Arduino

#define echo3 B01000000 // attach pin D2 Arduino
#define trig3 B10000000 //attach pin D3 Arduino

#define echo4 B000001 // attach pin D2 Arduino
#define trig4 B000010 //attach pin D3 Arduino
*/



#define L_PUMP    6   //Left pump
#define R_PUMP    3   //Right pump

#define MAX_PUMP  140  //Don't change this value! >140 can destroy the motor
#define TIMESTEP  20   //Control loop timer in ms
#define MAX_LEVEL 25 //Maximum safe level in tanks


#define echo1 30 // attach pin D2 Arduino
#define trig1 31 //attach pin D3 Arduino

#define echo2 32 // attach pin D2 Arduino
#define trig2 33 //attach pin D3 Arduino

#define echo3 34 // attach pin D2 Arduino
#define trig3 35 //attach pin D3 Arduino

#define echo4 36 // attach pin D2 Arduino
#define trig4 37 //attach pin D3 Arduino

const int BUTTON_RIGHT = 46;
const int BUTTON_LEFT = 47;
const int BUTTON_STOP = 48;




////_-------- PARAMETERS START ----------- ////





// Tank Inside Cross-Section Area for the Medium Outflow Ofrice Diameter (cm^2)
const double a = PI * 4.445 * 4.445 / 4;

// Tank 1 AND 2 Outlet Area  
const double o_1 = PI * 0.47625 * 0.47625 / 4; const double o_2 = PI * 0.47625 * 0.47625 / 4;
// Tank 3 AND 4 Outlet Area  
const double o_3 = PI * 0.47625 * 0.47625 / 4; const double o_4 = PI * 0.47625 * 0.47625 / 4;

//grav. const. (cm^3/s^2)
double g = 981;

// Pump Flow Constant (cm^3/s/V)
const double kp = 3.3;

// Sampling Time
const double Ts = TIMESTEP/1000;
///--------------------------------------------------------------------------------------------///
/// --- START --- TO BE CHANGED ///

//Offsets
const double h1_offset = 0.; //[cm]
const double h2_offset = 0.; //[cm]
const double h3_offset = 0.; //[cm]
const double h4_offset = 0.; //[cm]
//References
const double h1_ss = 15; //tank 1 level [cm]     
const double h2_ss = 15; //tank 2 level [cm]


//Flow ratios
double gamma1 = 0.72;
double gamma2 = 0.7;
const double h3_ss = h2_ss * (gamma1 * gamma1); //tank 3 level [cm]
const double h4_ss = h1_ss * (gamma2 * gamma2); //tank 4 level [cm]
//Voltages
const double v1_ss = 9.2606;
const double v2_ss = 9.2606;
///---------------- END --------------------------------------------------------------////



//constants T - MATRIX ENTRIES
const double T1 = a/o_1*sqrt(2*h1_ss/g);
const double T2 = a/o_2*sqrt(2*h2_ss/g);
const double T3 = a/o_3*sqrt(2*h3_ss/g);
const double T4 = a/o_4*sqrt(2*h4_ss/g);







//// --------- PARAMETERS END ----------- ////





















void read_sensors(double y[4]);
void write_actuators(unsigned int u[2], double y[4]);
void manual_control(double y[4], unsigned int u[2]);
void print_info(unsigned long currentMillis, double y[4], unsigned int u[2]);

extern void pi_controller(double y[4], unsigned int u[2]);
extern void pi_controller_dec(double y[4], unsigned int u[2]);
extern void LQR_controller(double y[4], unsigned int u[2]);
extern void LQR_int_controller(double y[4], unsigned int u[2]);


Matrix<4> KalmanFilter(const Matrix<4>& y, /*Matrix<4>& x_hat_new ,*/ Matrix<4>& x_hat, Matrix<2>& u, Matrix<4,4>& F, Matrix<4, 2>& B, Matrix<4,4>& P, Matrix<4, 4>& Q, Matrix<4, 4>& R, Matrix<4, 4>& K);
Matrix<4> KalmanFilter_DEBUG(const Matrix<4>& y, /*Matrix<4>& x_hat_new,*/ Matrix<4>& x_hat, Matrix<2>& u, Matrix<4,4>& F, Matrix<4, 2>& B, Matrix<4,4>& P, Matrix<4, 4>& Q, Matrix<4, 4>& R, Matrix<4, 4>& K);
double Kalman_Filter1(double y);
double Kalman_Filter2(double y);
double Kalman_Filter3(double y);
double Kalman_Filter4(double y);
long GET_MEASUREMENTS(const int& echo, const int& trig, long duration);


double ctrl_1;
double ctrl_2;

bool PI_CTRL = false;           // enter in PI control mode
bool PI_CTRL_DEC = false;       // enter in PI control decoupled mode
bool LQR_CTRL = false;          // enter in LQR control mode
bool LQR_INT_CTRL = false;      // enter in LQR integral control mode

//variables for PI controller
double error1 = 0;
double sumError1 = 0.0;
double error2 = 0;
double sumError2 = 0.0;

//variables for decoupled controller
double y_z1 = 0;
double y_z2 = 0;
double y_z1_old = 0;
double y_z2_old = 0;
int ctrl_1_old = 0;
int ctrl_2_old = 0;

//variables for LQR int controller
double sumDev1 = 0;
double sumDev2 = 0;

const double K_d2V = MAX_PUMP/12.58; // [-/V] Arduino digital output per volt to motor


double distances[4];
// defines variables
double distunf1;
long duration1; // variable for the duration of sound wave travel
double distance1; // variable for the distance measurement

double distunf2;
long duration2; // variable for the duration of sound wave travel
double distance2; // variable for the distance measurement

double distunf3;
long duration3; // variable for the duration of sound wave travel
double distance3; // variable for the distance measurement

double distunf4;
long duration4; // variable for the duration of sound wave travel
double distance4; // variable for the distance measurement

double FULLDISTANCE1;
double FULLDISTANCE2;
double FULLDISTANCE3;
double FULLDISTANCE4;


    Identity<4,4> I;
    //Matrix<4,4> A_ct;
    Matrix<4,4> A_ct = {(-1)/T1, 0, Ts/T3, 0,
            0, (-1)/T2, 0, Ts/T4,
            0, 0, (-1)/T3, 0,
            0, 0, 0, (-1)/T4};
    
    //Matrix<2,4> B_ct;
    Matrix<4,2> B_ct = {gamma1*kp/a, 0,
            0, gamma2*kp/a,
            0, (1-gamma2*kp)/a,
            (1-gamma1*kp)/a, 0};
    
    
    //Matrix<4,4> F;
    //Matrix<4,2> B;
    
    Matrix<4,4> F = {1, 0, 0.0044, 0,
            0, 1, 0, 0.0044,
            0, 0, 1, 0,
            0, 0, 0, 1};
    Matrix<4,2> B = B_ct * Ts;
    
    float q1 = 0.125;
    float q2 = 0.125;
    float q3 = 0.025;
    float q4 = 0.025;
    
    
    
   //Matrix<4,4> Q;
    Matrix<4,4> Q = {q1, 0., 0., 0.,
        0., q2, 0., 0.,
        0., 0., q3, 0.,
        0., 0., 0., q4};

    float r1 = 20.;
    float r2 = 20.;
    float r3 = 35.;
    float r4 = 35.;
    
    
   // Matrix<4,4> R;
    Matrix<4,4> R = {r1, 0., 0., 0.,
        0., r2, 0., 0.,
        0., 0., r3, 0.,
        0., 0., 0., r4};
    
    //Matrix<4,4> P; 
    Matrix<4,4> P = {1., 0., 0., 0.,
        0., 1., 0., 0.,
        0., 0., 1., 0.,
        0., 0., 0., 1.}; 
    
    //Matrix<4,4> K;
    Matrix<4,4> K = {0., 0., 0., 0.,
        0., 0., 0., 0.,
        0., 0., 0., 0.,
        0., 0., 0., 0.}; //It will immediately change => just reserve the space.


    Matrix<4,4> KK = {0.55, 0., 0., 0.,
        0., 0.04, 0., 0.,
        0., 0., 0.04, 0.,
        0., 0., 0., 0.04}; //It will immediately change => just reserve the space.
        
    Matrix<4> y_Vec = {0., 0., 0., 0.};
    Matrix<2> u_Vec = {0., 0.};

    //Matrix<4> y_est;
    Matrix<4> y_est = {distances[0], distances[1], distances[2], distances[3]}; //initial guess (= x_hat_old) since we have an observational Matrix == Identity
    
    static unsigned int u[2] = {0,0};
        
    double x[4] = {0, 0, 0, 0};



void setup() {
    
    //pinMode(L_PUMP, OUTPUT);
    //pinMode(R_PUMP, OUTPUT);

    pinMode(trig1, OUTPUT); // Sets the trig as an OUTPUT
    pinMode(echo1, INPUT); // Sets the echo as an INPUT
    pinMode(trig2, OUTPUT);
    pinMode(echo2, INPUT);
    pinMode(trig3, OUTPUT);
    pinMode(echo3, INPUT);
    pinMode(trig4, OUTPUT);
    pinMode(echo4, INPUT);
    pinMode(BUTTON_RIGHT, INPUT);
    pinMode(BUTTON_LEFT, INPUT);
    pinMode(BUTTON_STOP, INPUT);

    delay(200);
    distances[0] = 31.;
    delay(20);
    distances[1] = 31.2;
    delay(20);
    distances[2] = 31.3;
    delay(20);
    distances[3] = 31.2;
    delay(20);


    
    Serial.begin(115200); // // Serial Communication is starting with 115200 of baudrate speed
}




void loop() {
    
    
  
    while(1){
        
        static unsigned long previousMillis = 0;
        
        unsigned long currentMillis = millis();
        double tex[4];
        if (currentMillis - previousMillis >= TIMESTEP) {
            read_sensors(x);
            tex[0] = x[0];
            tex[1] = x[1];
            tex[2] = x[2];
            tex[3] = x[3];
            /*Serial.println(x[0]);
            Serial.println(x[1]);
            Serial.println(x[2]);
            Serial.println(x[3]);*/
            y_Vec = {distances[0] - x[0], distances[1] - x[1], distances[2] - x[2], distances[3] - x[3]};
            u_Vec = {float(u[0]), float(u[1])};
            //auto_control(y, u);
           
            y_est = KalmanFilter(y_Vec, y_est, u_Vec, F, B, P, Q, R, K);
            
            x[0] = y_est(0);
            x[1] = y_est(1);
            x[2] = y_est(2);
            x[3] = y_est(3);
            manual_control(x, u);

            if (PI_CTRL) {
               pi_controller(x, u);
            }
            if (PI_CTRL_DEC) {
              pi_controller_dec(x, u);
            }
            if (LQR_CTRL) {
              LQR_controller(x, u);
            }
            if (LQR_INT_CTRL) {
              LQR_int_controller(x, u);
            }
            }
        print_info(currentMillis, x, tex, u);
        write_actuators(u, x);
        previousMillis = currentMillis;
    }
}






long GET_MEASUREMENTS(const int& echo, const int& trig, long duration){

    digitalWrite(trig, LOW);
    delayMicroseconds(5);
    // Sets the trig1 HIGH (ACTIVE) for 20 microseconds
    digitalWrite(trig, HIGH);
    delayMicroseconds(20);
    digitalWrite(trig, LOW);
    // Reads the echoPin, returns the sound wave travel time in microseconds
    duration = pulseIn(echo, HIGH);
    // Calculating the distance
    return duration;

}


void read_sensors(double y[4]){
    y[0] = GET_MEASUREMENTS(echo1, trig1, duration1) * 0.034 / 2;
    y[1] = GET_MEASUREMENTS(echo2, trig2, duration2) * 0.034 / 2;
    y[2] = GET_MEASUREMENTS(echo3, trig3, duration3) * 0.034 / 2;
    y[3] = GET_MEASUREMENTS(echo4, trig4, duration4) * 0.034 / 2;
}


void print_info(unsigned long currentMillis, double y[4], double tex[4], unsigned int u[2]) {
    Serial.print(currentMillis);
    
    for (int i = 0; i < 4; ++i) {
        
        Serial.print("\t");
        Serial.print(distances[i] - tex[i]); // show unfiltered data (MAKE A BOOL AS SERIAL INPUT IF PEOPLE WANT TO SEE)
        
        Serial.print("\t");
        if(y[i] < 0){
            Serial.print(0);
        }
        else{
           Serial.print(y[i]);
        }
    }
    for (int i = 0; i < 2; ++i) {
        Serial.print("\t");
        Serial.print(u[i]);
    }

    Serial.print("\r\n");
}

void write_actuators(unsigned int u[2], double y[4])
{
    //Make actuators output safe
    for (int i = 0; i < 2; ++i) {
        if (u[i] < 0) {
            u[i] = 0;
            //Serial.println("Requested u is out of bounds!");
        }
        if (u[i] > MAX_PUMP) {
            u[i] = MAX_PUMP;
        }
    }
    if (y[0] > MAX_LEVEL || y[1] > MAX_LEVEL || y[2] > MAX_LEVEL || y[3] > MAX_LEVEL) {
        u[0] = 0;
        u[1] = 0;
    }

    //INVERT OUTPUTS HERE
    bool invert = false; //change to true to invert

    if (!invert) {
        analogWrite(L_PUMP, u[0]);
        analogWrite(R_PUMP, u[1]);
    } else {
        analogWrite(L_PUMP, u[1]);
        analogWrite(R_PUMP, u[0]);
    }
}


void manual_control(double y[4], unsigned int u[2]) {
    //Manual Mode for tests
    byte user_input;
/*    int stateLEFT = digitalRead(BUTTON_LEFT);
    int stateRIGHT = digitalRead(BUTTON_RIGHT);
    int stateSTOP = digitalRead(BUTTON_STOP);
    
    if (Serial.available()) {
        if(stateLEFT == 1){
          user_input = 'b';
        }
        else if (stateRIGHT == 1){
          user_input = 'n';
        }
        else if (stateSTOP == 1){
          user_input = 's';  
        }
        else {
          user_input = Serial.read();
        }*/
        user_input = Serial.read();
        switch (user_input) {
            //      case 'h':
            //        u[0] --;
            //        break;
            //      case 'j':
            //        u[0] ++;
            //        break;
            //      case 'k':
            //        u[1] --;
            //        break;
            //      case 'l':
            //        u[1] ++;
            //        break;
            case 's':
                u[0] = 0;
                u[1] = 0;
                PI_CTRL = false;
                PI_CTRL_DEC = false;
                LQR_CTRL = false;
                LQR_INT_CTRL = false;
                break;
            case 'a':
                u[0] = MAX_PUMP >> 1;
                u[1] = MAX_PUMP >> 1;
                break;
            case 'm':
                u[0] = MAX_PUMP;
                u[1] = MAX_PUMP;
                break;
            case 'b':
                u[0] = MAX_PUMP;
                break;
            case 'n':
                u[1] = MAX_PUMP;
                break;
            case 'p':
                PI_CTRL = true;
                error1 = 0;
                sumError1 = 0.0;
                error2 = 0;
                sumError2 = 0.0;
                break;
            case 'd':
                PI_CTRL_DEC = true;
                y_z1 = 0;
                y_z2 = 0;
                y_z1_old = 0;
                y_z2_old = 0;
                ctrl_1_old = 0;
                ctrl_2_old = 0;
                break;
            case 'l':
                LQR_CTRL = true;
                break;
            case 'q':
              LQR_INT_CTRL = true;
              sumDev1 = 0;
              sumDev2 = 0;
              break;
              }
        }






// If you're interested in how this KalmanFilter works and why it is set up as is, read the script

Matrix<4> KalmanFilter(const Matrix<4>& y, Matrix<4>& x_hat, Matrix<2>& u, Matrix<4,4>& F, Matrix<4, 2>& B, Matrix<4,4>& P, Matrix<4, 4>& Q, Matrix<4, 4>& R, Matrix<4, 4>& K){
    //1. ----------------- >x_hat enthält zu Beginn die Tankhöhen (31.7, 31.7, 31.7, 31.7)
    
    Matrix<4> x_hat_prediction = {0., 0., 0., 0.};
    
    /// --- START PREDICT --- ///
    x_hat_prediction = F*x_hat + B*u;
    P = F*P*(~F) + Q; //Tilde makes a Transpose -------- F is eye matrix...

    /// --- END   PREDICT --- ///
    
    Matrix<4,4> PR = P+R;
    //Matrix<4,4, Array<4,4, long long>> PR_inv = Invert(PR); //2. ------------- >Hier erhalte ich eine Matrix die folgend aussieht [a, 0, 0, 0][0, 0, ,0, 0][0, 0, 0, 0][0, 0, 0, 0] wobei a != 0
    float a = 1 / PR(0,0);float b = 1 / PR(1,1);float c = 1 / PR(2,2);float d = 1 / PR(3,3);
    Matrix<4,4> PR_inv = {a, 0., 0., 0.,
        0., b, 0., 0.,
        0., 0., c, 0.,
        0., 0., 0., d};
    /// --- START UPDATE --- ///
    K = P*PR_inv; // ------------- >3. Wegen 2. erhalte ich eine Matrix die folgend aussieht [a, 0, 0, 0][0, 0, ,0, 0][0, 0, 0, 0][0, 0, 0, 0] wobei a != 0

    P = (I - K)*P;//*IK_trans + K*R*(~K);
    
    x_hat = x_hat + K*(y - x_hat_prediction); //This will be returend //4. ----------- >Hier wird nun dadurch nur der erste Eintrag von x_hat verändert.
    
 
    /// --- END   UPDATE --- ///
    // Since the measurement Matrix is C = H = I => This means, that y_tilde = x_hat_new and the measurement
    return x_hat;
}



Matrix<4> KalmanFilter_DEBUG(const Matrix<4>& y, /*Matrix<4>& x_hat_new,*/ Matrix<4>& x_hat, Matrix<2>& u, Matrix<4,4>& F, Matrix<4, 2>& B, Matrix<4,4>& P, Matrix<4, 4>& Q, Matrix<4, 4>& R, Matrix<4, 4>& K){
    Matrix<4> x_hat_prediction = {0., 0., 0., 0.};
    /// --- START PREDICT --- ///
    Serial << "F:  " << F;
    Serial.println("");
    Serial << "B:  " << B;
    Serial.println("");
    Serial << "x_hat:  " << x_hat;
    Serial.println("");
    x_hat_prediction = F*x_hat;// + B*u;
    Serial << "x_hat_prediction:  " << x_hat_prediction;
    Serial.println("");
    P = F*P*(~F) + Q; //Tilde makes a Transpose
    Serial << "Pfirst:  " << P;
    Serial.println("");
    /// --- END   PREDICT --- ///
    
    Matrix<4,4> PR = P+R;
    Serial << "PR:  " << PR;
    Serial.println("");
    float a = 1 / PR(0,0);float b = 1 / PR(1,1);float c = 1 / PR(2,2);float d = 1 / PR(3,3);
    Matrix<4,4> PR_inv = {a, 0., 0., 0.,
        0., b, 0., 0.,
        0., 0., c, 0.,
        0., 0., 0., d};
    Serial << "PR_inv:  " << PR_inv;
    
    Serial.println("");
    /// --- START UPDATE --- ///
    K = P*PR_inv;
    Serial << "K:  " << K;
    Serial.println("");
    //Matrix<4,4> IK = I-K;
    //Matrix<4,4> IK_trans = ~(IK);

    x_hat += K*(y - x_hat_prediction); //This will be returend
    P = (I - K)*P;//*IK_trans + K*R*(~K);
    Serial << "Psecond:  " << P;
    Serial.println("");
    Serial.println("");
    Serial.println("");
    Serial.println("");
    
    /// --- END   UPDATE --- ///
    // Since the measurement Matrix is C = H = I => This means, that y_tilde = x_hat_new and the measurement
    return x_hat;
}
