%% start xpc and setup model parameters

%%%init xpc target
%TargetInit()

%%%setup tank-parameters and sensors


% Coupled Tanks' system nomenclature:
% a         Tank Inside Cross-Section Area           (cm^2)
% o_i       Tank i Outlet Area                       (cm^2)
% Do_i      Outflow Orifice Diameter of tank i       (cm)
% Do_s      Small Outflow Orifice Diameter           (cm)
% Do_m      Medium Outflow Orifice Diameter          (cm)
% Do_l      Large Outflow Orifice Diameter           (cm)
% g         Gravity Constant                         (cm/s^2)
% kp        Pump Flow Constant                       (cm^3/s/V)
% Ts        Sampling Time                            (s)
% K_L       Tank Water Level Sensor Sensitivity      (cm/V)
% L_MAX     Tank Height (Maximum Water Level)        (cm)
% Vp_MAX    Pump Maximum Voltage                     (V)
% VMAX_UPM  UPM Maximum Output Voltage               (V)
% IMAX_UPM  UPM Maximum Output Current               (A)

% UPM-2405 Maximum Output Voltage (V) and Current (A)
global VMAX_UPM IMAX_UPM
% safety water level limits (cm)
global L_MAX
% Pressure Sensor Sensitivities (cm/V)
global K_L

% %%%Diameters
% % Tank Inside Diameter (cm)
% Da = 4.445; % = 1.75 inch
% % Small Outflow Orifice Diameter (cm)
% Do_s = 0.31750; % = 1/8 in
% % Medium Outflow Orifice Diameter (cm)
% Do_m = 0.47625;  % = 3/16 in
% % Large Outflow Orifice Diameter (cm)
% Do_l = 0.55563; % = 7/32 in

%%%Calculate the system's areas

% Tank Inside Cross-Section Area for the Medium Outflow Ofrice Diameter (cm^2)
a = pi * 4.445^2 / 4;
% Tank 1 Outlet Area  
o_1 = pi * 0.47625^2 / 4;
% Tank 2 Outlet Area  
o_2 = pi * 0.47625^2 / 4;
% Tank 3 Outlet Area  
o_3 = pi * 0.47625^2 / 4;
% Tank 4 Outlet Area  
o_4 = pi * 0.47625^2 / 4;

%%%Constants

% Gravitational Constant on Earth (cm/s^2)
g = 981;
% Pump Flow Constant (cm^3/s/V)
kp = 3.3;
% Sampling Time
Ts=0.01;
% Tank 1 Water Level Sensor Sensitivity (cm/V)
K_L = 25 / 4.08;
% Tank Height (i.e. Maximum Water Level) (cm)
L_MAX = 27; %mq-->reduced this a bit, for safety
% Pump Maximum Continuous Voltage(V)
Vp_MAX = 12;
% UPM-2405 Maximum Output Voltage (V) and Current (A)
VMAX_UPM = 22;
% UPM-2405 Maximum Output Voltage (V) and Current (A)
IMAX_UPM = 5;

%%%sensors

% Cable Gain used
K_CABLE = 5;
% Digital-to-Analog Maximum Voltage (V); for MultiQ cards set to 10
VMAX_DAC = 10;
% Low-pass filter time constant 
tau_t = 1 / ( 2 * pi * 2 );






